@@ -1,6 +1,6 @@
 // Protocol Buffers - Google's data interchange format
 // Copyright 2008 Google Inc.  All rights reserved.
-// https://developers.google.com/protocol-buffers/
+// http://code.google.com/p/protobuf/
 //
 // Redistribution and use in source and binary forms, with or without
 // modification, are permitted provided that the following conditions are
@@ -35,23 +35,14 @@
 #ifndef GOOGLE_PROTOBUF_COMPILER_JAVA_FIELD_H__
 #define GOOGLE_PROTOBUF_COMPILER_JAVA_FIELD_H__
 
-#include <map>
-#include <memory>
 #include <string>
-
 #include <google/protobuf/stubs/common.h>
 #include <google/protobuf/descriptor.h>
 
 namespace google {
 namespace protobuf {
-  namespace compiler {
-    namespace java {
-      class Context;                // context.h
-      class ClassNameResolver;      // name_resolver.h
-    }
-  }
   namespace io {
-    class Printer;                  // printer.h
+    class Printer;             // printer.h
   }
 }
 
@@ -59,10 +50,10 @@ namespace protobuf {
 namespace compiler {
 namespace java {
 
-class ImmutableFieldGenerator {
+class FieldGenerator {
  public:
-  ImmutableFieldGenerator() {}
-  virtual ~ImmutableFieldGenerator();
+  FieldGenerator() {}
+  virtual ~FieldGenerator();
 
   virtual int GetNumBitsForMessage() const = 0;
   virtual int GetNumBitsForBuilder() const = 0;
@@ -87,73 +78,29 @@ class ImmutableFieldGenerator {
   virtual string GetBoxedType() const = 0;
 
  private:
-  GOOGLE_DISALLOW_EVIL_CONSTRUCTORS(ImmutableFieldGenerator);
+  GOOGLE_DISALLOW_EVIL_CONSTRUCTORS(FieldGenerator);
 };
 
-
 // Convenience class which constructs FieldGenerators for a Descriptor.
-template<typename FieldGeneratorType>
 class FieldGeneratorMap {
  public:
-  explicit FieldGeneratorMap(const Descriptor* descriptor,
-                             Context* context);
+  explicit FieldGeneratorMap(const Descriptor* descriptor);
   ~FieldGeneratorMap();
 
-  const FieldGeneratorType& get(const FieldDescriptor* field) const;
+  const FieldGenerator& get(const FieldDescriptor* field) const;
+  const FieldGenerator& get_extension(int index) const;
 
  private:
   const Descriptor* descriptor_;
-  Context* context_;
-  ClassNameResolver* name_resolver_;
-  scoped_array<scoped_ptr<FieldGeneratorType> > field_generators_;
+  scoped_array<scoped_ptr<FieldGenerator> > field_generators_;
+  scoped_array<scoped_ptr<FieldGenerator> > extension_generators_;
+
+  static FieldGenerator* MakeGenerator(const FieldDescriptor* field,
+      int messageBitIndex, int builderBitIndex);
 
   GOOGLE_DISALLOW_EVIL_CONSTRUCTORS(FieldGeneratorMap);
 };
 
-template<typename FieldGeneratorType>
-inline const FieldGeneratorType&
-FieldGeneratorMap<FieldGeneratorType>::get(const FieldDescriptor* field) const {
-  GOOGLE_CHECK_EQ(field->containing_type(), descriptor_);
-  return *field_generators_[field->index()];
-}
-
-// Instantiate template for mutable and immutable maps.
-template<>
-FieldGeneratorMap<ImmutableFieldGenerator>::
-FieldGeneratorMap(const Descriptor* descriptor,
-                  Context* context);
-
-template<>
-FieldGeneratorMap<ImmutableFieldGenerator>::~FieldGeneratorMap();
-
-
-// Field information used in FieldGeneartors.
-struct FieldGeneratorInfo {
-  string name;
-  string capitalized_name;
-  string disambiguated_reason;
-};
-
-// Oneof information used in OneofFieldGeneartors.
-struct OneofGeneratorInfo {
-  string name;
-  string capitalized_name;
-};
-
-// Set some common variables used in variable FieldGenerators.
-void SetCommonFieldVariables(const FieldDescriptor* descriptor,
-                             const FieldGeneratorInfo* info,
-                             map<string, string>* variables);
-
-// Set some common oneof variables used in OneofFieldGenerators.
-void SetCommonOneofVariables(const FieldDescriptor* descriptor,
-                             const OneofGeneratorInfo* info,
-                             map<string, string>* variables);
-
-// Print useful comments before a field's accessors.
-void PrintExtraFieldInfo(const map<string, string>& variables,
-                         io::Printer* printer);
-
 }  // namespace java
 }  // namespace compiler
 }  // namespace protobuf
