@@ -33,9 +33,9 @@
 //  Sanjay Ghemawat, Jeff Dean, and others.
 
 #include <algorithm>
+#include <google/protobuf/descriptor.pb.h>
 #include <google/protobuf/generated_message_reflection.h>
 #include <google/protobuf/descriptor.h>
-#include <google/protobuf/descriptor.pb.h>
 #include <google/protobuf/repeated_field.h>
 #include <google/protobuf/extension_set.h>
 #include <google/protobuf/generated_message_util.h>
@@ -45,8 +45,6 @@ namespace google {
 namespace protobuf {
 namespace internal {
 
-namespace { const string kEmptyString; }
-
 int StringSpaceUsedExcludingSelf(const string& str) {
   const void* start = &str;
   const void* end = &str + 1;
@@ -69,7 +67,6 @@ bool ParseNamedEnum(const EnumDescriptor* descriptor,
 }
 
 const string& NameOfEnum(const EnumDescriptor* descriptor, int value) {
-  static string kEmptyString;
   const EnumValueDescriptor* d = descriptor->FindValueByNumber(value);
   return (d == NULL ? kEmptyString : d->name());
 }
@@ -239,8 +236,13 @@ int GeneratedMessageReflection::SpaceUsed(const Message& message) const {
 #undef HANDLE_TYPE
 
         case FieldDescriptor::CPPTYPE_STRING:
-            total_size += GetRaw<RepeatedPtrField<string> >(message, field)
-                            .SpaceUsedExcludingSelf();
+          switch (field->options().ctype()) {
+            default:  // TODO(kenton):  Support other string reps.
+            case FieldOptions::STRING:
+              total_size += GetRaw<RepeatedPtrField<string> >(message, field)
+                              .SpaceUsedExcludingSelf();
+              break;
+          }
           break;
 
         case FieldDescriptor::CPPTYPE_MESSAGE:
@@ -265,18 +267,24 @@ int GeneratedMessageReflection::SpaceUsed(const Message& message) const {
           break;
 
         case FieldDescriptor::CPPTYPE_STRING: {
-            const string* ptr = GetField<const string*>(message, field);
+          switch (field->options().ctype()) {
+            default:  // TODO(kenton):  Support other string reps.
+            case FieldOptions::STRING: {
+              const string* ptr = GetField<const string*>(message, field);
 
-            // Initially, the string points to the default value stored in
-            // the prototype. Only count the string if it has been changed
-            // from the default value.
-            const string* default_ptr = DefaultRaw<const string*>(field);
+              // Initially, the string points to the default value stored in
+              // the prototype. Only count the string if it has been changed
+              // from the default value.
+              const string* default_ptr = DefaultRaw<const string*>(field);
 
-            if (ptr != default_ptr) {
-              // string fields are represented by just a pointer, so also
-              // include sizeof(string) as well.
-              total_size += sizeof(*ptr) + StringSpaceUsedExcludingSelf(*ptr);
+              if (ptr != default_ptr) {
+                // string fields are represented by just a pointer, so also
+                // include sizeof(string) as well.
+                total_size += sizeof(*ptr) + StringSpaceUsedExcludingSelf(*ptr);
+              }
+              break;
             }
+          }
           break;
         }
 
@@ -313,7 +321,7 @@ void GeneratedMessageReflection::Swap(
        "descriptor.";
   GOOGLE_CHECK_EQ(message2->GetReflection(), this)
     << "Second argument to Swap() (of type \""
-    << message1->GetDescriptor()->full_name()
+    << message2->GetDescriptor()->full_name()
     << "\") is not compatible with this reflection object (which is for type \""
     << descriptor_->full_name()
     << "\").  Note that the exact same class is required; not just the same "
@@ -324,7 +332,7 @@ void GeneratedMessageReflection::Swap(
   int has_bits_size = (descriptor_->field_count() + 31) / 32;
 
   for (int i = 0; i < has_bits_size; i++) {
-    swap(has_bits1[i], has_bits2[i]);
+    std::swap(has_bits1[i], has_bits2[i]);
   }
 
   for (int i = 0; i < descriptor_->field_count(); i++) {
@@ -360,8 +368,8 @@ void GeneratedMessageReflection::Swap(
       switch (field->cpp_type()) {
 #define SWAP_VALUES(CPPTYPE, TYPE)                                           \
         case FieldDescriptor::CPPTYPE_##CPPTYPE:                             \
-          swap(*MutableRaw<TYPE>(message1, field),                           \
-               *MutableRaw<TYPE>(message2, field));                          \
+          std::swap(*MutableRaw<TYPE>(message1, field),                      \
+                    *MutableRaw<TYPE>(message2, field));                     \
           break;
 
           SWAP_VALUES(INT32 , int32 );
@@ -372,12 +380,20 @@ void GeneratedMessageReflection::Swap(
           SWAP_VALUES(DOUBLE, double);
           SWAP_VALUES(BOOL  , bool  );
           SWAP_VALUES(ENUM  , int   );
-          SWAP_VALUES(MESSAGE, Message*);
 #undef SWAP_VALUES
+        case FieldDescriptor::CPPTYPE_MESSAGE:
+          std::swap(*MutableRaw<Message*>(message1, field),
+                    *MutableRaw<Message*>(message2, field));
+          break;
 
         case FieldDescriptor::CPPTYPE_STRING:
-            swap(*MutableRaw<string*>(message1, field),
-                 *MutableRaw<string*>(message2, field));
+          switch (field->options().ctype()) {
+            default:  // TODO(kenton):  Support other string reps.
+            case FieldOptions::STRING:
+              std::swap(*MutableRaw<string*>(message1, field),
+                        *MutableRaw<string*>(message2, field));
+              break;
+          }
           break;
 
         default:
@@ -473,15 +489,20 @@ void GeneratedMessageReflection::ClearField(
           break;
 
         case FieldDescriptor::CPPTYPE_STRING: {
-            const string* default_ptr = DefaultRaw<const string*>(field);
-            string** value = MutableRaw<string*>(message, field);
-            if (*value != default_ptr) {
-              if (field->has_default_value()) {
-                (*value)->assign(field->default_value_string());
-              } else {
-                (*value)->clear();
+          switch (field->options().ctype()) {
+            default:  // TODO(kenton):  Support other string reps.
+            case FieldOptions::STRING:
+              const string* default_ptr = DefaultRaw<const string*>(field);
+              string** value = MutableRaw<string*>(message, field);
+              if (*value != default_ptr) {
+                if (field->has_default_value()) {
+                  (*value)->assign(field->default_value_string());
+                } else {
+                  (*value)->clear();
+                }
               }
-            }
+              break;
+          }
           break;
         }
 
@@ -508,7 +529,12 @@ void GeneratedMessageReflection::ClearField(
 #undef HANDLE_TYPE
 
       case FieldDescriptor::CPPTYPE_STRING: {
-          MutableRaw<RepeatedPtrField<string> >(message, field)->Clear();
+        switch (field->options().ctype()) {
+          default:  // TODO(kenton):  Support other string reps.
+          case FieldOptions::STRING:
+            MutableRaw<RepeatedPtrField<string> >(message, field)->Clear();
+            break;
+        }
         break;
       }
 
@@ -549,7 +575,12 @@ void GeneratedMessageReflection::RemoveLast(
 #undef HANDLE_TYPE
 
       case FieldDescriptor::CPPTYPE_STRING:
-          MutableRaw<RepeatedPtrField<string> >(message, field)->RemoveLast();
+        switch (field->options().ctype()) {
+          default:  // TODO(kenton):  Support other string reps.
+          case FieldOptions::STRING:
+            MutableRaw<RepeatedPtrField<string> >(message, field)->RemoveLast();
+            break;
+        }
         break;
 
       case FieldDescriptor::CPPTYPE_MESSAGE:
@@ -560,6 +591,20 @@ void GeneratedMessageReflection::RemoveLast(
   }
 }
 
+Message* GeneratedMessageReflection::ReleaseLast(
+    Message* message,
+    const FieldDescriptor* field) const {
+  USAGE_CHECK_ALL(ReleaseLast, REPEATED, MESSAGE);
+
+  if (field->is_extension()) {
+    return static_cast<Message*>(
+        MutableExtensionSet(message)->ReleaseLast(field->number()));
+  } else {
+    return MutableRaw<RepeatedPtrFieldBase>(message, field)
+        ->ReleaseLast<GenericTypeHandler<Message> >();
+  }
+}
+
 void GeneratedMessageReflection::SwapElements(
     Message* message,
     const FieldDescriptor* field,
@@ -658,7 +703,7 @@ void GeneratedMessageReflection::ListFields(
     USAGE_CHECK_ALL(Set##TYPENAME, SINGULAR, CPPTYPE);                       \
     if (field->is_extension()) {                                             \
       return MutableExtensionSet(message)->Set##TYPENAME(                    \
-        field->number(), field->type(), value);                              \
+        field->number(), field->type(), value, field);                       \
     } else {                                                                 \
       SetField<TYPE>(message, field, value);                                 \
     }                                                                        \
@@ -694,7 +739,8 @@ void GeneratedMessageReflection::ListFields(
     USAGE_CHECK_ALL(Add##TYPENAME, REPEATED, CPPTYPE);                       \
     if (field->is_extension()) {                                             \
       MutableExtensionSet(message)->Add##TYPENAME(                           \
-        field->number(), field->type(), field->options().packed(), value);   \
+        field->number(), field->type(), field->options().packed(), value,    \
+        field);                                                              \
     } else {                                                                 \
       AddField<TYPE>(message, field, value);                                 \
     }                                                                        \
@@ -718,7 +764,14 @@ string GeneratedMessageReflection::GetString(
     return GetExtensionSet(message).GetString(field->number(),
                                               field->default_value_string());
   } else {
-    return *GetField<const string*>(message, field);
+    switch (field->options().ctype()) {
+      default:  // TODO(kenton):  Support other string reps.
+      case FieldOptions::STRING:
+        return *GetField<const string*>(message, field);
+    }
+
+    GOOGLE_LOG(FATAL) << "Can't get here.";
+    return kEmptyString;  // Make compiler happy.
   }
 }
 
@@ -730,7 +783,14 @@ const string& GeneratedMessageReflection::GetStringReference(
     return GetExtensionSet(message).GetString(field->number(),
                                               field->default_value_string());
   } else {
-    return *GetField<const string*>(message, field);
+    switch (field->options().ctype()) {
+      default:  // TODO(kenton):  Support other string reps.
+      case FieldOptions::STRING:
+        return *GetField<const string*>(message, field);
+    }
+
+    GOOGLE_LOG(FATAL) << "Can't get here.";
+    return kEmptyString;  // Make compiler happy.
   }
 }
 
@@ -741,13 +801,19 @@ void GeneratedMessageReflection::SetString(
   USAGE_CHECK_ALL(SetString, SINGULAR, STRING);
   if (field->is_extension()) {
     return MutableExtensionSet(message)->SetString(field->number(),
-                                                   field->type(), value);
+                                                   field->type(), value, field);
   } else {
-    string** ptr = MutableField<string*>(message, field);
-    if (*ptr == DefaultRaw<const string*>(field)) {
-      *ptr = new string(value);
-    } else {
-      (*ptr)->assign(value);
+    switch (field->options().ctype()) {
+      default:  // TODO(kenton):  Support other string reps.
+      case FieldOptions::STRING: {
+        string** ptr = MutableField<string*>(message, field);
+        if (*ptr == DefaultRaw<const string*>(field)) {
+          *ptr = new string(value);
+        } else {
+          (*ptr)->assign(value);
+        }
+        break;
+      }
     }
   }
 }
@@ -759,7 +825,14 @@ string GeneratedMessageReflection::GetRepeatedString(
   if (field->is_extension()) {
     return GetExtensionSet(message).GetRepeatedString(field->number(), index);
   } else {
-    return GetRepeatedPtrField<string>(message, field, index);
+    switch (field->options().ctype()) {
+      default:  // TODO(kenton):  Support other string reps.
+      case FieldOptions::STRING:
+        return GetRepeatedPtrField<string>(message, field, index);
+    }
+
+    GOOGLE_LOG(FATAL) << "Can't get here.";
+    return kEmptyString;  // Make compiler happy.
   }
 }
 
@@ -770,7 +843,14 @@ const string& GeneratedMessageReflection::GetRepeatedStringReference(
   if (field->is_extension()) {
     return GetExtensionSet(message).GetRepeatedString(field->number(), index);
   } else {
-    return GetRepeatedPtrField<string>(message, field, index);
+    switch (field->options().ctype()) {
+      default:  // TODO(kenton):  Support other string reps.
+      case FieldOptions::STRING:
+        return GetRepeatedPtrField<string>(message, field, index);
+    }
+
+    GOOGLE_LOG(FATAL) << "Can't get here.";
+    return kEmptyString;  // Make compiler happy.
   }
 }
 
@@ -783,7 +863,12 @@ void GeneratedMessageReflection::SetRepeatedString(
     MutableExtensionSet(message)->SetRepeatedString(
       field->number(), index, value);
   } else {
-    *MutableRepeatedField<string>(message, field, index) = value;
+    switch (field->options().ctype()) {
+      default:  // TODO(kenton):  Support other string reps.
+      case FieldOptions::STRING:
+        *MutableRepeatedField<string>(message, field, index) = value;
+        break;
+    }
   }
 }
 
@@ -794,9 +879,14 @@ void GeneratedMessageReflection::AddString(
   USAGE_CHECK_ALL(AddString, REPEATED, STRING);
   if (field->is_extension()) {
     MutableExtensionSet(message)->AddString(field->number(),
-                                            field->type(), value);
+                                            field->type(), value, field);
   } else {
-    *AddField<string>(message, field) = value;
+    switch (field->options().ctype()) {
+      default:  // TODO(kenton):  Support other string reps.
+      case FieldOptions::STRING:
+        *AddField<string>(message, field) = value;
+        break;
+    }
   }
 }
 
@@ -816,7 +906,9 @@ const EnumValueDescriptor* GeneratedMessageReflection::GetEnum(
   }
   const EnumValueDescriptor* result =
     field->enum_type()->FindValueByNumber(value);
-  GOOGLE_CHECK(result != NULL);
+  GOOGLE_CHECK(result != NULL) << "Value " << value << " is not valid for field "
+                        << field->full_name() << " of type "
+                        << field->enum_type()->full_name() << ".";
   return result;
 }
 
@@ -828,7 +920,7 @@ void GeneratedMessageReflection::SetEnum(
 
   if (field->is_extension()) {
     MutableExtensionSet(message)->SetEnum(field->number(), field->type(),
-                                          value->number());
+                                          value->number(), field);
   } else {
     SetField<int>(message, field, value->number());
   }
@@ -846,7 +938,9 @@ const EnumValueDescriptor* GeneratedMessageReflection::GetRepeatedEnum(
   }
   const EnumValueDescriptor* result =
     field->enum_type()->FindValueByNumber(value);
-  GOOGLE_CHECK(result != NULL);
+  GOOGLE_CHECK(result != NULL) << "Value " << value << " is not valid for field "
+                        << field->full_name() << " of type "
+                        << field->enum_type()->full_name() << ".";
   return result;
 }
 
@@ -874,7 +968,7 @@ void GeneratedMessageReflection::AddEnum(
   if (field->is_extension()) {
     MutableExtensionSet(message)->AddEnum(field->number(), field->type(),
                                           field->options().packed(),
-                                          value->number());
+                                          value->number(), field);
   } else {
     AddField<int>(message, field, value->number());
   }
@@ -883,16 +977,19 @@ void GeneratedMessageReflection::AddEnum(
 // -------------------------------------------------------------------
 
 const Message& GeneratedMessageReflection::GetMessage(
-    const Message& message, const FieldDescriptor* field) const {
+    const Message& message, const FieldDescriptor* field,
+    MessageFactory* factory) const {
   USAGE_CHECK_ALL(GetMessage, SINGULAR, MESSAGE);
 
+  if (factory == NULL) factory = message_factory_;
+
   if (field->is_extension()) {
     return static_cast<const Message&>(
-      GetExtensionSet(message).GetMessage(field->number(),
-                                          field->message_type(),
-                                          message_factory_));
+        GetExtensionSet(message).GetMessage(
+          field->number(), field->message_type(), factory));
   } else {
-    const Message* result = GetRaw<const Message*>(message, field);
+    const Message* result;
+    result = GetRaw<const Message*>(message, field);
     if (result == NULL) {
       result = DefaultRaw<const Message*>(field);
     }
@@ -901,22 +998,44 @@ const Message& GeneratedMessageReflection::GetMessage(
 }
 
 Message* GeneratedMessageReflection::MutableMessage(
-    Message* message, const FieldDescriptor* field) const {
+    Message* message, const FieldDescriptor* field,
+    MessageFactory* factory) const {
   USAGE_CHECK_ALL(MutableMessage, SINGULAR, MESSAGE);
 
+  if (factory == NULL) factory = message_factory_;
+
   if (field->is_extension()) {
     return static_cast<Message*>(
-        MutableExtensionSet(message)->MutableMessage(field->number(),
-                                                     field->type(),
-                                                     field->message_type(),
-                                                     message_factory_));
+        MutableExtensionSet(message)->MutableMessage(field, factory));
   } else {
-    Message** result = MutableField<Message*>(message, field);
-    if (*result == NULL) {
+    Message* result;
+    Message** result_holder = MutableField<Message*>(message, field);
+    if (*result_holder == NULL) {
       const Message* default_message = DefaultRaw<const Message*>(field);
-      *result = default_message->New();
+      *result_holder = default_message->New();
     }
-    return *result;
+    result = *result_holder;
+    return result;
+  }
+}
+
+Message* GeneratedMessageReflection::ReleaseMessage(
+    Message* message,
+    const FieldDescriptor* field,
+    MessageFactory* factory) const {
+  USAGE_CHECK_ALL(ReleaseMessage, SINGULAR, MESSAGE);
+
+  if (factory == NULL) factory = message_factory_;
+
+  if (field->is_extension()) {
+    return static_cast<Message*>(
+        MutableExtensionSet(message)->ReleaseMessage(field, factory));
+  } else {
+    ClearBit(message, field);
+    Message** result = MutableRaw<Message*>(message, field);
+    Message* ret = *result;
+    *result = NULL;
+    return ret;
   }
 }
 
@@ -948,26 +1067,26 @@ Message* GeneratedMessageReflection::MutableRepeatedMessage(
 }
 
 Message* GeneratedMessageReflection::AddMessage(
-    Message* message, const FieldDescriptor* field) const {
+    Message* message, const FieldDescriptor* field,
+    MessageFactory* factory) const {
   USAGE_CHECK_ALL(AddMessage, REPEATED, MESSAGE);
 
+  if (factory == NULL) factory = message_factory_;
+
   if (field->is_extension()) {
     return static_cast<Message*>(
-        MutableExtensionSet(message)->AddMessage(field->number(),
-                                                 field->type(),
-                                                 field->message_type(),
-                                                 message_factory_));
+        MutableExtensionSet(message)->AddMessage(field, factory));
   } else {
     // We can't use AddField<Message>() because RepeatedPtrFieldBase doesn't
     // know how to allocate one.
     RepeatedPtrFieldBase* repeated =
-      MutableRaw<RepeatedPtrFieldBase>(message, field);
+        MutableRaw<RepeatedPtrFieldBase>(message, field);
     Message* result = repeated->AddFromCleared<GenericTypeHandler<Message> >();
     if (result == NULL) {
       // We must allocate a new object.
       const Message* prototype;
       if (repeated->size() == 0) {
-        prototype = message_factory_->GetPrototype(field->message_type());
+        prototype = factory->GetPrototype(field->message_type());
       } else {
         prototype = &repeated->Get<GenericTypeHandler<Message> >(0);
       }
@@ -978,7 +1097,26 @@ Message* GeneratedMessageReflection::AddMessage(
   }
 }
 
-// -------------------------------------------------------------------
+void* GeneratedMessageReflection::MutableRawRepeatedField(
+    Message* message, const FieldDescriptor* field,
+    FieldDescriptor::CppType cpptype,
+    int ctype, const Descriptor* desc) const {
+  USAGE_CHECK_REPEATED("MutableRawRepeatedField");
+  if (field->cpp_type() != cpptype)
+    ReportReflectionUsageTypeError(descriptor_,
+        field, "MutableRawRepeatedField", cpptype);
+  if (ctype >= 0)
+    GOOGLE_CHECK_EQ(field->options().ctype(), ctype) << "subtype mismatch";
+  if (desc != NULL)
+    GOOGLE_CHECK_EQ(field->message_type(), desc) << "wrong submessage type";
+  if (field->is_extension())
+    return MutableExtensionSet(message)->MutableRawRepeatedField(
+        field->number());
+  else
+    return reinterpret_cast<uint8*>(message) + offsets_[field->index()];
+}
+
+// -----------------------------------------------------------------------------
 
 const FieldDescriptor* GeneratedMessageReflection::FindKnownExtensionByName(
     const string& name) const {
@@ -1110,7 +1248,7 @@ inline Type* GeneratedMessageReflection::MutableField(
 }
 
 template <typename Type>
-inline Type GeneratedMessageReflection::GetRepeatedField(
+inline const Type& GeneratedMessageReflection::GetRepeatedField(
     const Message& message, const FieldDescriptor* field, int index) const {
   return GetRaw<RepeatedField<Type> >(message, field).Get(index);
 }
@@ -1138,7 +1276,7 @@ inline Type* GeneratedMessageReflection::MutableRepeatedField(
 
 template <typename Type>
 inline void GeneratedMessageReflection::AddField(
-    Message* message, const FieldDescriptor* field, Type value) const {
+    Message* message, const FieldDescriptor* field, const Type& value) const {
   MutableRaw<RepeatedField<Type> >(message, field)->Add(value);
 }
 
