@@ -48,12 +48,11 @@
 #include <google/protobuf/unknown_field_set.h>
 
 
+
 namespace google {
 namespace protobuf {
 namespace internal {
 
-using internal::WireFormatLite;
-
 namespace {
 
 // This function turns out to be convenient when using some macros later.
@@ -239,8 +238,6 @@ void WireFormat::SerializeUnknownMessageSetItems(
     // The only unknown fields that are allowed to exist in a MessageSet are
     // messages, which are length-delimited.
     if (field.type() == UnknownField::TYPE_LENGTH_DELIMITED) {
-      const string& data = field.length_delimited();
-
       // Start group.
       output->WriteVarint32(WireFormatLite::kMessageSetItemStartTag);
 
@@ -250,8 +247,7 @@ void WireFormat::SerializeUnknownMessageSetItems(
 
       // Write message.
       output->WriteVarint32(WireFormatLite::kMessageSetMessageTag);
-      output->WriteVarint32(data.size());
-      output->WriteString(data);
+      field.SerializeLengthDelimitedNoTag(output);
 
       // End group.
       output->WriteVarint32(WireFormatLite::kMessageSetItemEndTag);
@@ -268,8 +264,6 @@ uint8* WireFormat::SerializeUnknownMessageSetItemsToArray(
     // The only unknown fields that are allowed to exist in a MessageSet are
     // messages, which are length-delimited.
     if (field.type() == UnknownField::TYPE_LENGTH_DELIMITED) {
-      const string& data = field.length_delimited();
-
       // Start group.
       target = io::CodedOutputStream::WriteTagToArray(
           WireFormatLite::kMessageSetItemStartTag, target);
@@ -283,8 +277,7 @@ uint8* WireFormat::SerializeUnknownMessageSetItemsToArray(
       // Write message.
       target = io::CodedOutputStream::WriteTagToArray(
           WireFormatLite::kMessageSetMessageTag, target);
-      target = io::CodedOutputStream::WriteVarint32ToArray(data.size(), target);
-      target = io::CodedOutputStream::WriteStringToArray(data, target);
+      target = field.SerializeLengthDelimitedNoTagToArray(target);
 
       // End group.
       target = io::CodedOutputStream::WriteTagToArray(
@@ -354,9 +347,10 @@ int WireFormat::ComputeUnknownMessageSetItemsSize(
     if (field.type() == UnknownField::TYPE_LENGTH_DELIMITED) {
       size += WireFormatLite::kMessageSetItemTagsSize;
       size += io::CodedOutputStream::VarintSize32(field.number());
-      size += io::CodedOutputStream::VarintSize32(
-        field.length_delimited().size());
-      size += field.length_delimited().size();
+
+      int field_size = field.GetLengthDelimitedSize();
+      size += io::CodedOutputStream::VarintSize32(field_size);
+      size += field_size;
     }
   }
 
@@ -391,7 +385,12 @@ bool WireFormat::ParseAndMergePartial(io::CodedInputStream* input,
 
       // If that failed, check if the field is an extension.
       if (field == NULL && descriptor->IsExtensionNumber(field_number)) {
-        field = message_reflection->FindKnownExtensionByNumber(field_number);
+        if (input->GetExtensionPool() == NULL) {
+          field = message_reflection->FindKnownExtensionByNumber(field_number);
+        } else {
+          field = input->GetExtensionPool()
+                       ->FindExtensionByNumber(descriptor, field_number);
+        }
       }
 
       // If that failed, but we're a MessageSet, and this is the tag for a
@@ -419,52 +418,67 @@ bool WireFormat::ParseAndMergeField(
     io::CodedInputStream* input) {
   const Reflection* message_reflection = message->GetReflection();
 
-  if (field == NULL ||
-      WireFormatLite::GetTagWireType(tag) != WireTypeForField(field)) {
-    // We don't recognize this field.  Either the field number is unknown
-    // or the wire type doesn't match.  Put it in our unknown field set.
-    return SkipField(input, tag,
-                     message_reflection->MutableUnknownFields(message));
+  enum { UNKNOWN, NORMAL_FORMAT, PACKED_FORMAT } value_format;
+
+  if (field == NULL) {
+    value_format = UNKNOWN;
+  } else if (WireFormatLite::GetTagWireType(tag) ==
+             WireTypeForFieldType(field->type())) {
+    value_format = NORMAL_FORMAT;
+  } else if (field->is_packable() &&
+             WireFormatLite::GetTagWireType(tag) ==
+             WireFormatLite::WIRETYPE_LENGTH_DELIMITED) {
+    value_format = PACKED_FORMAT;
+  } else {
+    // We don't recognize this field. Either the field number is unknown
+    // or the wire type doesn't match. Put it in our unknown field set.
+    value_format = UNKNOWN;
   }
 
-  if (field->options().packed()) {
+  if (value_format == UNKNOWN) {
+    return SkipField(input, tag,
+                     message_reflection->MutableUnknownFields(message));
+  } else if (value_format == PACKED_FORMAT) {
     uint32 length;
     if (!input->ReadVarint32(&length)) return false;
     io::CodedInputStream::Limit limit = input->PushLimit(length);
 
     switch (field->type()) {
-#define HANDLE_PACKED_TYPE(TYPE, TYPE_METHOD, CPPTYPE, CPPTYPE_METHOD)         \
+#define HANDLE_PACKED_TYPE(TYPE, CPPTYPE, CPPTYPE_METHOD)                      \
       case FieldDescriptor::TYPE_##TYPE: {                                     \
         while (input->BytesUntilLimit() > 0) {                                 \
           CPPTYPE value;                                                       \
-          if (!WireFormatLite::Read##TYPE_METHOD(input, &value)) return false; \
+          if (!WireFormatLite::ReadPrimitive<                                  \
+                CPPTYPE, WireFormatLite::TYPE_##TYPE>(input, &value))          \
+            return false;                                                      \
           message_reflection->Add##CPPTYPE_METHOD(message, field, value);      \
         }                                                                      \
         break;                                                                 \
       }
 
-      HANDLE_PACKED_TYPE( INT32,  Int32,  int32,  Int32)
-      HANDLE_PACKED_TYPE( INT64,  Int64,  int64,  Int64)
-      HANDLE_PACKED_TYPE(SINT32, SInt32,  int32,  Int32)
-      HANDLE_PACKED_TYPE(SINT64, SInt64,  int64,  Int64)
-      HANDLE_PACKED_TYPE(UINT32, UInt32, uint32, UInt32)
-      HANDLE_PACKED_TYPE(UINT64, UInt64, uint64, UInt64)
+      HANDLE_PACKED_TYPE( INT32,  int32,  Int32)
+      HANDLE_PACKED_TYPE( INT64,  int64,  Int64)
+      HANDLE_PACKED_TYPE(SINT32,  int32,  Int32)
+      HANDLE_PACKED_TYPE(SINT64,  int64,  Int64)
+      HANDLE_PACKED_TYPE(UINT32, uint32, UInt32)
+      HANDLE_PACKED_TYPE(UINT64, uint64, UInt64)
 
-      HANDLE_PACKED_TYPE( FIXED32,  Fixed32, uint32, UInt32)
-      HANDLE_PACKED_TYPE( FIXED64,  Fixed64, uint64, UInt64)
-      HANDLE_PACKED_TYPE(SFIXED32, SFixed32,  int32,  Int32)
-      HANDLE_PACKED_TYPE(SFIXED64, SFixed64,  int64,  Int64)
+      HANDLE_PACKED_TYPE( FIXED32, uint32, UInt32)
+      HANDLE_PACKED_TYPE( FIXED64, uint64, UInt64)
+      HANDLE_PACKED_TYPE(SFIXED32,  int32,  Int32)
+      HANDLE_PACKED_TYPE(SFIXED64,  int64,  Int64)
 
-      HANDLE_PACKED_TYPE(FLOAT , Float , float , Float )
-      HANDLE_PACKED_TYPE(DOUBLE, Double, double, Double)
+      HANDLE_PACKED_TYPE(FLOAT , float , Float )
+      HANDLE_PACKED_TYPE(DOUBLE, double, Double)
 
-      HANDLE_PACKED_TYPE(BOOL, Bool, bool, Bool)
+      HANDLE_PACKED_TYPE(BOOL, bool, Bool)
 #undef HANDLE_PACKED_TYPE
 
       case FieldDescriptor::TYPE_ENUM: {
         while (input->BytesUntilLimit() > 0) {
           int value;
-          if (!WireFormatLite::ReadEnum(input, &value)) return false;
+          if (!WireFormatLite::ReadPrimitive<int, WireFormatLite::TYPE_ENUM>(
+                  input, &value)) return false;
           const EnumValueDescriptor* enum_value =
               field->enum_type()->FindValueByNumber(value);
           if (enum_value != NULL) {
@@ -487,11 +501,14 @@ bool WireFormat::ParseAndMergeField(
 
     input->PopLimit(limit);
   } else {
+    // Non-packed value (value_format == NORMAL_FORMAT)
     switch (field->type()) {
-#define HANDLE_TYPE(TYPE, TYPE_METHOD, CPPTYPE, CPPTYPE_METHOD)               \
+#define HANDLE_TYPE(TYPE, CPPTYPE, CPPTYPE_METHOD)                            \
       case FieldDescriptor::TYPE_##TYPE: {                                    \
         CPPTYPE value;                                                        \
-        if (!WireFormatLite::Read##TYPE_METHOD(input, &value)) return false;  \
+        if (!WireFormatLite::ReadPrimitive<                                   \
+                CPPTYPE, WireFormatLite::TYPE_##TYPE>(input, &value))         \
+          return false;                                                       \
         if (field->is_repeated()) {                                           \
           message_reflection->Add##CPPTYPE_METHOD(message, field, value);     \
         } else {                                                              \
@@ -500,31 +517,28 @@ bool WireFormat::ParseAndMergeField(
         break;                                                                \
       }
 
-      HANDLE_TYPE( INT32,  Int32,  int32,  Int32)
-      HANDLE_TYPE( INT64,  Int64,  int64,  Int64)
-      HANDLE_TYPE(SINT32, SInt32,  int32,  Int32)
-      HANDLE_TYPE(SINT64, SInt64,  int64,  Int64)
-      HANDLE_TYPE(UINT32, UInt32, uint32, UInt32)
-      HANDLE_TYPE(UINT64, UInt64, uint64, UInt64)
+      HANDLE_TYPE( INT32,  int32,  Int32)
+      HANDLE_TYPE( INT64,  int64,  Int64)
+      HANDLE_TYPE(SINT32,  int32,  Int32)
+      HANDLE_TYPE(SINT64,  int64,  Int64)
+      HANDLE_TYPE(UINT32, uint32, UInt32)
+      HANDLE_TYPE(UINT64, uint64, UInt64)
 
-      HANDLE_TYPE( FIXED32,  Fixed32, uint32, UInt32)
-      HANDLE_TYPE( FIXED64,  Fixed64, uint64, UInt64)
-      HANDLE_TYPE(SFIXED32, SFixed32,  int32,  Int32)
-      HANDLE_TYPE(SFIXED64, SFixed64,  int64,  Int64)
+      HANDLE_TYPE( FIXED32, uint32, UInt32)
+      HANDLE_TYPE( FIXED64, uint64, UInt64)
+      HANDLE_TYPE(SFIXED32,  int32,  Int32)
+      HANDLE_TYPE(SFIXED64,  int64,  Int64)
 
-      HANDLE_TYPE(FLOAT , Float , float , Float )
-      HANDLE_TYPE(DOUBLE, Double, double, Double)
-
-      HANDLE_TYPE(BOOL, Bool, bool, Bool)
-
-      HANDLE_TYPE(STRING, String, string, String)
-      HANDLE_TYPE(BYTES, Bytes, string, String)
+      HANDLE_TYPE(FLOAT , float , Float )
+      HANDLE_TYPE(DOUBLE, double, Double)
 
+      HANDLE_TYPE(BOOL, bool, Bool)
 #undef HANDLE_TYPE
 
       case FieldDescriptor::TYPE_ENUM: {
         int value;
-        if (!WireFormatLite::ReadEnum(input, &value)) return false;
+        if (!WireFormatLite::ReadPrimitive<int, WireFormatLite::TYPE_ENUM>(
+                input, &value)) return false;
         const EnumValueDescriptor* enum_value =
           field->enum_type()->FindValueByNumber(value);
         if (enum_value != NULL) {
@@ -544,13 +558,38 @@ bool WireFormat::ParseAndMergeField(
         break;
       }
 
+      // Handle strings separately so that we can optimize the ctype=CORD case.
+      case FieldDescriptor::TYPE_STRING: {
+        string value;
+        if (!WireFormatLite::ReadString(input, &value)) return false;
+        VerifyUTF8String(value.data(), value.length(), PARSE);
+        if (field->is_repeated()) {
+          message_reflection->AddString(message, field, value);
+        } else {
+          message_reflection->SetString(message, field, value);
+        }
+        break;
+      }
+
+      case FieldDescriptor::TYPE_BYTES: {
+        string value;
+        if (!WireFormatLite::ReadBytes(input, &value)) return false;
+        if (field->is_repeated()) {
+          message_reflection->AddString(message, field, value);
+        } else {
+          message_reflection->SetString(message, field, value);
+        }
+        break;
+      }
 
       case FieldDescriptor::TYPE_GROUP: {
         Message* sub_message;
         if (field->is_repeated()) {
-          sub_message = message_reflection->AddMessage(message, field);
+          sub_message = message_reflection->AddMessage(
+              message, field, input->GetExtensionFactory());
         } else {
-          sub_message = message_reflection->MutableMessage(message, field);
+          sub_message = message_reflection->MutableMessage(
+              message, field, input->GetExtensionFactory());
         }
 
         if (!WireFormatLite::ReadGroup(WireFormatLite::GetTagFieldNumber(tag),
@@ -562,9 +601,11 @@ bool WireFormat::ParseAndMergeField(
       case FieldDescriptor::TYPE_MESSAGE: {
         Message* sub_message;
         if (field->is_repeated()) {
-          sub_message = message_reflection->AddMessage(message, field);
+          sub_message = message_reflection->AddMessage(
+              message, field, input->GetExtensionFactory());
         } else {
-          sub_message = message_reflection->MutableMessage(message, field);
+          sub_message = message_reflection->MutableMessage(
+              message, field, input->GetExtensionFactory());
         }
 
         if (!WireFormatLite::ReadMessage(input, sub_message)) return false;
@@ -595,10 +636,7 @@ bool WireFormat::ParseAndMergeMessageSetItem(
   const FieldDescriptor* field = NULL;
 
   // If we see message data before the type_id, we'll append it to this so
-  // we can parse it later.  This will probably never happen in practice,
-  // as no MessageSet encoder I know of writes the message before the type ID.
-  // But, it's technically valid so we should allow it.
-  // TODO(kenton):  Use a Cord instead?  Do I care?
+  // we can parse it later.
   string message_data;
 
   while (true) {
@@ -636,7 +674,10 @@ bool WireFormat::ParseAndMergeMessageSetItem(
           uint32 length;
           if (!input->ReadVarint32(&length)) return false;
           if (!input->ReadString(&temp, length)) return false;
-          message_data.append(temp);
+          io::StringOutputStream output_stream(&message_data);
+          io::CodedOutputStream coded_output(&output_stream);
+          coded_output.WriteVarint32(length);
+          coded_output.WriteString(temp);
         } else {
           // Already saw type_id, so we can parse this directly.
           if (!ParseAndMergeField(fake_tag, field, message, input)) {
@@ -782,23 +823,23 @@ void WireFormat::SerializeFieldWithCachedSizes(
       // Handle strings separately so that we can get string references
       // instead of copying.
       case FieldDescriptor::TYPE_STRING: {
-          string scratch;
-          const string& value = field->is_repeated() ?
-            message_reflection->GetRepeatedStringReference(
-              message, field, j, &scratch) :
-            message_reflection->GetStringReference(message, field, &scratch);
-          VerifyUTF8String(value.data(), value.length(), SERIALIZE);
-          WireFormatLite::WriteString(field->number(), value, output);
+        string scratch;
+        const string& value = field->is_repeated() ?
+          message_reflection->GetRepeatedStringReference(
+            message, field, j, &scratch) :
+          message_reflection->GetStringReference(message, field, &scratch);
+        VerifyUTF8String(value.data(), value.length(), SERIALIZE);
+        WireFormatLite::WriteString(field->number(), value, output);
         break;
       }
 
       case FieldDescriptor::TYPE_BYTES: {
-          string scratch;
-          const string& value = field->is_repeated() ?
-            message_reflection->GetRepeatedStringReference(
-              message, field, j, &scratch) :
-            message_reflection->GetStringReference(message, field, &scratch);
-          WireFormatLite::WriteBytes(field->number(), value, output);
+        string scratch;
+        const string& value = field->is_repeated() ?
+          message_reflection->GetRepeatedStringReference(
+            message, field, j, &scratch) :
+          message_reflection->GetStringReference(message, field, &scratch);
+        WireFormatLite::WriteBytes(field->number(), value, output);
         break;
       }
     }
@@ -961,14 +1002,14 @@ int WireFormat::FieldDataOnlyByteSize(
     // instead of copying.
     case FieldDescriptor::TYPE_STRING:
     case FieldDescriptor::TYPE_BYTES: {
-        for (int j = 0; j < count; j++) {
-          string scratch;
-          const string& value = field->is_repeated() ?
-            message_reflection->GetRepeatedStringReference(
-              message, field, j, &scratch) :
-            message_reflection->GetStringReference(message, field, &scratch);
-          data_size += WireFormatLite::StringSize(value);
-        }
+      for (int j = 0; j < count; j++) {
+        string scratch;
+        const string& value = field->is_repeated() ?
+          message_reflection->GetRepeatedStringReference(
+            message, field, j, &scratch) :
+          message_reflection->GetStringReference(message, field, &scratch);
+        data_size += WireFormatLite::StringSize(value);
+      }
       break;
     }
   }
@@ -1009,10 +1050,10 @@ void WireFormat::VerifyUTF8StringFallback(const char* data,
         break;
       // no default case: have the compiler warn if a case is not covered.
     }
-    GOOGLE_LOG(ERROR) << "Encountered string containing invalid UTF-8 data while "
+    GOOGLE_LOG(ERROR) << "String field contains invalid UTF-8 data when "
                << operation_str
-               << " protocol buffer. Strings must contain only UTF-8; "
-                  "use the 'bytes' type for raw bytes.";
+               << " a protocol buffer. Use the 'bytes' type if you intend to "
+                  "send raw bytes.";
   }
 }
 
